 /***********************************************
***
***  eeeeee eeeee eeeee eeeee eeeeee eeeeee
***  8    8 8     8     8     8   88 8    8
***  8e   8 8     8     8     8    8 8    8
***  88   8 8eee  8eee  8e    8    8 8e   8
***  88   8 8e    8e    88    8    8 88   8
***  88   e 88    88    88    8    e 88   8
***  88eee8 88eee 88    88eee 88eee8 88   8
*** ----------------------------------------
***         defcon-cc.dyndns.org
***
***        Coded by Crashdemon....
***         HAVE A LOT OF PHUN!
***
***			 	gps laptimer
***        
************************************************
************************************************
***   Programm: gps laptimer                 ***
***   Autor: Patrick Langosch                ***
***   erstellt: 24.02.13                     ***
***   Version: 0.3                          ***
***********************************************/
/***********************************************
***
***	 Anschlubelegung ATMEGA16
***
***  Pin 	Description		Connected with
***  -----------------------------------------
***  1		PB0(XCK/TO)	    -
***  2		PB1(T1)			-
***  3      PB2(INT2/AIN0)	-
***  4		PB3(OC0/AIN1)	-
***  5		PB4(SS)			-
***  6		PB5(MOSI)		-
***  7		PB6(MISO)		-
***  8		PB7(SCK)		-
***  9		RESET			-
***	 10		VCC				+5V
***  11		GND				Ground
***  12		XTAL2			4Mhz Quarz
***  13		XTAL1			4Mhz Quarz
***  14		PD0(RXD)		GPS(TXD)
***  15 	PD1(TXD)		GPS(RXD)
***  16		PD2(INT0)		-
***  17		PD3(INT1)		-
***  18		PD4(OC1B)		-
***  19		PD5(OC1A)		-
***  20		PD6(ICP1)		-
***  21		PD7(OC2)		-
***  22		PC0(SCL)		-
***  23		PC1(SDA)		-
***  24		PC2(TCK)		JTAG-ICE
***  25		PC3(TMS)		JTAG-ICE
***  26		PC4(TDO)		JTAG-ICE
***  27		PC5(TDI)		JTAG-ICE
***  28		PC6(TOSC1)		-
***  29		PC7(TOSC1)		-
***	 30		AVCC			+5V			
***  31		GND				Ground
***  32		AREF			+3,3V
***  33		PA7(ADC7)		-				
***  34		PA6(ADC6)		LCD BackGrnd
***  35		PA5(ADC5)		LCD EN
***	 36		PA4(ADC4)		LCD RS
***  37		PA3(ADC3)		LCD DB7
***  38		PA2(ADC2)		LCD DB6
***  39		PA1(ADC1)		LCD DB5
***  40		PA0(ADC0)		LCD DB4
***
***********************************************/
/***********************************************
***
***  Changelog:
***  - 0.1 / some major bugs
***  - 0.2 / running version
***  - 0.3 / some hardware workaround 
***          + lcd backlight ctrl
***	   
***********************************************/

#include <math.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <avr/io.h>
#include <util/delay.h>
#include <avr/pgmspace.h>
#include <avr/interrupt.h>

#define TRUE 1
#define FALSE 0

#define _STR(a) #a
#define STR(a) _STR(a)

#define VERSION 0.3
#define F_CPU 4000000UL // Systemtakt in Hz

// DEBUG
#define DEBUG

// GPS
#define PATTERN_SIZE 3
#define BUFFER_SIZE 2 
#define EARTH_RADIUS_IN_METERS 6372797.560856

// LCD
#define CLEAR_DISPLAY 0x01 // Display Inhalt komplett mit Null berschreiben
#define CURSOR_HOME 0x02 // Coursor an Anfangsposition setzen
#define LCD_PORT PORTA // Port an dem das LCD angeschlossen ist
#define LCD_DDR DDRA // Datenpins an denen das LCD angeschlossen ist
#define LCD_RS PA4 // Ausgang an dem der RS-Pin des LCD hngt
#define LCD_EN PA5 // Ausgang an dem der EN-Pin des LCD hngt
#define LCD_BACKGRND PA6 // Schaltet Hintergrundbeleuchtung

// USART
#define BAUD 38400UL
#define UBRR_VALUE (F_CPU/8/BAUD-1)
#define USART_DDR DDRD
#define USART_RX_PIN PD0
#define USART_TX_PIN PD1

// structure holds minutes, seconds and centi-seconds 
struct myTime 
{
	uint8_t min;
	uint8_t sec;
	uint8_t csec;
};

// structure holds hours, minutes, seconds and centi-seconds 
struct myUTCTime 
{
	uint8_t hour;
	uint8_t min;
	uint8_t sec;
	uint8_t csec;
};

// GPSData structure
struct GPSData
{
	char rawLatitude[12]; // GGA
	char rawLongitude[12]; // GGA
	char satellites[3]; // GGA
	float latitude;	// GGA
	float longitude; // GGA
	float heading; // VTG
	float speed; // VTG
	struct myUTCTime time; // GGA
};

// Global Variables
uint8_t noHeadingFlag = 1;
volatile uint16_t overflow_ticks = 0;
char cBuffer[20] = " ";

float topSpeed = 0.0;
char buffer[BUFFER_SIZE * 80];

struct GPSData homePos;
struct GPSData actGPSPos;

struct myTime lap; // the laptime of the last round
struct myTime act;	// actual laptime
struct myTime show; // the displayed time
struct myTime best; // besttime

static const char patternVTG[] = "VTG";
static const char patternGGA[] = "GGA";
static const char patternRMC[] = "RMC";

volatile int8_t isVTG = -1;
volatile int8_t isGGA = -1;

static char headingText[] =
	"  N\0NNE\0 NE\0ENE\0  E\0ESE\0 SE\0SSE\0"
	"  S\0SSW\0 SW\0WSW\0  W\0WNW\0 NW\0NNW";

// some strings in program memory
const char INFO_1[] PROGMEM = "DefCon";
const char INFO_2[] PROGMEM = "GPS Laptimer";
const char INFO_3[] PROGMEM = "> Waiting for GPS";
const char INFO_4[] PROGMEM = "> GPS sig. valid?";
const char INFO_5[] PROGMEM = "> Set NMEA record";
const char INFO_6[] PROGMEM = "> Set speed thr.";


// Prototypendeklaration
// Initialisierungen
void initUSART(void);
void initTimer1(void);
void initGPS(void);
void LCD_init(void); 
void initStruct(struct myTime*);

// GPS
void reset_time(void);
void splitTicks(uint16_t, volatile struct myTime*);
void runningMode(void);
void waitForStartLine(void);
void stoppingMode(void);
char* nWord(char*, uint8_t);
float wordToFloat(char *);
float wordToRad(char*);
float wordToTime(char*);
float haversine(struct GPSData*, struct GPSData*);
void parseGPSData(struct GPSData*);
uint8_t isLapNewBestTime(struct myTime*, struct myTime*);

// LCD
void LCD_clear(void);
void LCD_home(void);
void LCD_data(unsigned char);
void LCD_string(const char*);
void LCD_intro(void);
void LCD_enable(void);
void LCD_command(unsigned char);
void LCD_set_cursor(uint8_t, uint8_t);
void LCD_showTime(struct myTime*);
void LCD_showLaptime(void);
void LCD_showSpeed(struct GPSData*);
void LCD_showUTCTime(struct myUTCTime* time);
void LCD_ctrlBacklight(uint8_t);

#ifdef DEBUG
void LCD_showDebug(struct GPSData*);
#endif

// USART
void initUSART(void);
void USART_Send(const char*);
char USART_Read_Byte(void);
uint8_t USART_recvACK(char*, uint8_t);
uint8_t USART_isValidGPS(char*, uint8_t);

/******** ISR of the USART(RX) ********/
ISR(USART_RXC_vect)
{
	char data; // temporary memory to serve interrupt reason
	char* ptrMsgBegin; 
	static uint8_t index = BUFFER_SIZE; 
	static uint8_t startMsg = 0; // marks beginning of a message
	static uint16_t timerOffset = 0; // save current timer value
	
	data = UDR; // Save the char, to serve the interrupt reason

	// begin of the message
	if(data == '$')
	{
		timerOffset = TCNT1;

		// buffer can handle up to two messages, one VTG and one GGA
		for(index = 0; index < BUFFER_SIZE; index++)
		{
			if((index != isVTG) && (index != isGGA))
			{
				startMsg = (uint8_t)(index * 80);
				break;		
			}	
		}
	}

	// free space?
	if(index < BUFFER_SIZE)
	{
		// end of the message
		if(data == '\r')
		{
			// mark the end of the message
			buffer[startMsg] = '\0';
			ptrMsgBegin = (char*)(&buffer[index * 80 + 3]);
			
			// no VTG message saved in buffer and message is VTG
			if((isVTG < 0) && (strncmp(ptrMsgBegin, patternVTG, PATTERN_SIZE) == 0))
			{
				// save the pos. of VTG msg. in buffer
				isVTG = index;
			}	

			else if((isGGA < 0) && (strncmp(ptrMsgBegin, patternGGA, PATTERN_SIZE) == 0))
			{
				TCNT1 = -5000 + timerOffset; // correct timer value
				TIMSK |= (1 << TOIE1); // enable timer1 overflow interrupt

				// save the pos. of GGA msg. in buffer
				isGGA = index;
			}

			index = BUFFER_SIZE;
		}

		else
		{
			buffer[startMsg++] = data;	
		}
	}
}

/******** ISR des Overflow von Timer1 ********/
ISR(TIMER1_OVF_vect)
{
	TCNT1 = -5000; // overflow every 10ms
	overflow_ticks++;
}

/********************************** MAIN ******************************************/
int main(void)
{
	cli(); // disable global interrupts

	initStruct(&act); 
	initStruct(&show);
	initStruct(&lap);
	initStruct(&best);

	// initialization - right order is important
	LCD_init();
	LCD_intro();
	initUSART();
	initGPS();
	initTimer1();
	
	sei(); // enable global interrupts
	LCD_ctrlBacklight(FALSE);

	waitForStartLine(); // wait for user setting start line

	/********************************** LOOP ******************************************/
	while(1)
	{
		runningMode();
		stoppingMode();
	}
}

void initStruct(struct myTime* temp)
{
	temp->csec = 0;
	temp->sec = 0;
	temp->min = 0;
}

/*************************************** GPS ***********************************************/
void initGPS(void)
{
	LCD_clear();

	// waiting for GPS 
	strcpy_P(cBuffer, INFO_3);
	LCD_set_cursor(0, 1);
	LCD_string(cBuffer);

	// wait for a byte to appear over the uart
	while(!(UCSRA & (1 << RXC)));

	LCD_set_cursor(0, 1);
	LCD_data(' ');
	_delay_ms(500);

	// GPS sig. valid?
	strcpy_P(cBuffer, INFO_4);
	LCD_set_cursor(0, 2);
	LCD_string(cBuffer);

	USART_Send("$PMTK314,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*29\r\n"); // only RMC
	while(!USART_isValidGPS((char*)patternRMC, PATTERN_SIZE));

	LCD_set_cursor(0, 2);
	LCD_data(' ');
	_delay_ms(500);

	// set NMEA record
	strcpy_P(cBuffer, INFO_5);
	LCD_set_cursor(0, 3);
	LCD_string(cBuffer);

	USART_Send("$PMTK314,0,0,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*28\r\n"); // only VTG and GGA

	LCD_set_cursor(0, 3);
	LCD_data(' ');
	_delay_ms(500);

	// set speed thr.
	strcpy_P(cBuffer, INFO_6);
	LCD_set_cursor(0, 4);
	LCD_string(cBuffer);

	USART_Send("$PMTK397,0.6*3B\r\n"); // set nav speed threshold to 0,6 m/s --> 2,16 km/h

	// enable RX-ISR 
	UCSRB |= (1 << RXCIE);

	LCD_clear();
}

void waitForStartLine(void)
{
	/********* LCD *********/
	LCD_home();
	LCD_string("TIME:");
	
	LCD_set_cursor(0, 2);
	LCD_string("SPEED:");

	LCD_set_cursor(0, 3);
	LCD_string("HEADING:");

	LCD_set_cursor(0, 4);
	LCD_string("PRESS TO SET LINE");

	#ifdef DEBUG
		LCD_clear();
	#endif 

	while(1)
	{
		// wait for consistent GGA and VTG data
		if(isVTG >= 0 && isGGA >= 0)		
		{	
			// fit nav data into structure
			parseGPSData(&actGPSPos);

			#ifdef DEBUG
				LCD_showDebug(&actGPSPos);
			#endif

			#ifndef DEBUG
				// time
				LCD_set_cursor(9, 1);
				LCD_showUTCTime(&actGPSPos.time);

				// speed
				LCD_set_cursor(12, 2);
				LCD_showSpeed(&actGPSPos);
			
				// heading		
				LCD_set_cursor(17, 3);	
				LCD_string(&headingText[((int)(actGPSPos.heading) * 2 / 45) * 4]);	
			#endif 

			// accept new messages
			isVTG = -1;
			isGGA = -1;
		}

		// button "start line" pressed
		if(!(PIND & (1 << PD6)))
		{
			reset_time(); // begin to count
			homePos = actGPSPos; // save actual pos. as start line

			// if we are not moving, heading thru start point still needs to be worked out
			noHeadingFlag = !(actGPSPos.speed > 2.0);
			
			break; // escape from loop
		}
	}
}

void runningMode(void)
{
	uint8_t lap_complete = 0;
	uint8_t temp = 0;

	float distance = 0.0;
	float diffHeading = 0.0;
	float prevHomeDistance = 0.0;

	char gettingCloser = 0;

	LCD_clear();

	// get distance to home position from actual position
	prevHomeDistance = haversine(&homePos, &actGPSPos);

	while(1)
	{
		splitTicks(overflow_ticks, &act); // get time

		// wait for consistent GGA and VTG data
		if(isVTG >= 0 && isGGA >= 0)
		{
			// fit nav data into structure
			parseGPSData(&actGPSPos);

			// get distance to home position from actual position
			distance = haversine(&homePos, &actGPSPos);

			// speed
			LCD_set_cursor(12, 3);
			LCD_showSpeed(&actGPSPos);

			// are we moving?
			if(noHeadingFlag) // nope, slower then 2 km/h
			{
				// don't yet know what direction (heading) we pass the start finish line
				if(distance < prevHomeDistance)
				{
					// actual distance from start line is smaller then before
					// make actual distance new previous distance
					prevHomeDistance = distance;
				}

				else // far away from finish line (>prevHomeDistance)
				{
					homePos.heading = actGPSPos.heading; // save heading information
					noHeadingFlag = 0; // now we have heading information, disable flag
				}
			}

			else // yep faster then 2 km/h
			{
				// check if we are getting closer to the home position
				if(distance < prevHomeDistance)
				{
					temp = 1;
				}
				
				else
				{
					temp = 0;
				}

				gettingCloser = (gettingCloser << 1) | temp;

				prevHomeDistance = distance;

				// get the difference from heading at home postion to heading at actual position
				diffHeading = actGPSPos.heading - homePos.heading;
				
				// are we driving in the wrong direction?
				if(diffHeading > 180)
				{
					diffHeading = diffHeading - 360;
				}

	
				if((gettingCloser & 0x03) == 0x02 && (prevHomeDistance < 5.0) && (diffHeading > -30 && diffHeading < 30))
				{
					// got within 5 meters of the start point, heading in the right direction (+-30). 
					// call this a lap 					
					lap = act; // save current time
					overflow_ticks = 0; // reset counter
					lap_complete = 1; // round is finished

					// calculate fraction of laptime at finish line
					// comming soon...

					// is this laptime a new besttime?
					if(isLapNewBestTime(&best, &lap))
					{
						best = lap;
					}
				}
			}

			LCD_set_cursor(16, 4);
			if(distance < 10) LCD_data(' ');
			if(distance < 100) LCD_data(' ');
			if(distance < 1000) LCD_data(' ');
			LCD_string(itoa(distance, cBuffer, 10));

			// slower the 2km/h?
			if(actGPSPos.speed < 2.0)
			{				
				break; // leave runningMode, show total
			}

			if(lap_complete != 0)
			{
				if(act.sec < 5)
				{
					show = lap;
				}

				else
				{
					show = act;
					lap_complete = 0;
				}
			}

			else
			{
				show = act;
			}

			// reset!!
			isVTG = -1;
			isGGA = -1;
		}

		LCD_showLaptime();
	}
}

void stoppingMode(void)
{
	struct GPSData stopPos;

	stopPos = actGPSPos;

	LCD_clear();
	LCD_home();
	LCD_string("TOTAL");

	// show top speed
	LCD_set_cursor(0, 2);
	LCD_string("TOPSPEED:");

	LCD_set_cursor(12, 2);
	if(topSpeed < 100) LCD_data(' ');
	if(topSpeed < 10) LCD_data(' ');
	LCD_string(utoa((uint8_t)topSpeed, cBuffer, 10));
	LCD_string(" km/h");

	LCD_set_cursor(0, 3);
	LCD_string("BESTTIME:");
	LCD_set_cursor(12, 3);
	LCD_showTime(&best);

	/********* LCD *********/

	while(1)
	{
		// wait for consistent GGA and VTG data
		if(isVTG >= 0 && isGGA >= 0)
		{
			// fit nav data into structure
			parseGPSData(&actGPSPos);
			
			// more than 5 meters away from stopping position
			if(haversine(&stopPos, &actGPSPos) > 5.0)
			//if(actGPSPos.speed > 5.0)
			{
				break; // moving again
			} 	

			// reset!!
			isVTG = -1;
			isGGA = -1;
		}
	}
}

uint8_t isLapNewBestTime(struct myTime* besttime, struct myTime* laptime)
{
	uint32_t sumBest = 0;
	uint32_t sumLap = 0;

	// besttime contains nothing
	if(besttime->min == 0 && besttime->sec == 0 && besttime->csec == 0)
	{
		return(1);
	}

	else
	{
		// sum of the besttime
		sumBest  = besttime->min * 6000;
		sumBest += besttime->sec * 100;
		sumBest += besttime->csec;
		
		// sum of the laptime	
		sumLap  = laptime->min * 6000;
		sumLap += laptime->sec * 100;
		sumLap += laptime->csec;
		
		// laptime is smaller the besttime
		if(sumLap < sumBest)
		{
			return(1);
		}	
	}

	return(0);
}

float haversine(struct GPSData *A, struct GPSData *B)
{
	float dLat = 0.0;
	float dLon = 0.0;
	float distance = 0.0;

	// diff. from point A to B divided by two
	dLat = (B->latitude - A->latitude) / 2.0; 
	dLon = (B->longitude - A->longitude) / 2.0;

	distance = 2.0 * EARTH_RADIUS_IN_METERS * asin(sqrt(pow(sin(dLat), 2) + pow(sin(dLon), 2) * cos(B->latitude) * cos(A->latitude)));

	return(distance);
}

void reset_time(void)
{
	overflow_ticks = 0; // reset counter

	// Alle Member auf Null setzen
	act.min = 0;
	act.sec = 0;
	act.csec = 0;

	show = act; // Aktuelle Zeit anzeigen
}

void splitTicks(uint16_t ticks, volatile struct myTime* result)
{
	uint16_t centiSec = 0;
	uint16_t seconds = 0;

	centiSec = ticks; // centi-seconds (10ms)
	seconds = centiSec / 100; // calculate seconds

	result->csec = centiSec  % 100; // residual of seconds --> centi-seconds
	result->min = seconds  / 60; // calculate minutes
	result->sec = seconds  % 60; // residual of minutes --> seconds
}

void parseGPSData(struct GPSData *ptrGPS)
{
	char *ptr;
	char *temp_ptr;
	char *UTCbegin;

	char hour[3];
	char min[3];
	char sec[3];
	char csec[3];
		
	// Parse VTG sentence for heading and speed
	// heading
	ptr = &buffer[isVTG * 80];
	ptrGPS->heading = wordToFloat(nWord(ptr, 1));

	// speed
	ptrGPS->speed = wordToFloat(nWord(ptr, 7));
	
	// Parse GGA sentence for position & time
	ptr = &buffer[isGGA * 80];

	UTCbegin = nWord(ptr, 1);

	strncpy(hour, UTCbegin, 2);	
	strncpy(min, UTCbegin+2, 2);
	strncpy(sec, UTCbegin+4, 2);
	strncpy(csec, UTCbegin+7, 2);	
	
	hour[2] = '\0';
	min[2] = '\0';
	sec[2] = '\0';
	csec[2] = '\0';	
	
	ptrGPS->time.hour = atoi(hour);	
	ptrGPS->time.min = atoi(min);
	ptrGPS->time.sec = atoi(sec);
	ptrGPS->time.csec = atoi(csec);

	// satellites used
	temp_ptr = nWord(ptr, 7);
	
	// check if character at ptr is digit
	if(*(temp_ptr+1) >= '0' && *(temp_ptr+1) <= '9')
	{
		// 10 to 14 satellites
		strncpy(ptrGPS->satellites, temp_ptr, 2);
		ptrGPS->satellites[2] = '\0';
	}

	else
	{
		// 0 to 9 satellites
		ptrGPS->satellites[0] = ' ';
		strncpy(ptrGPS->satellites+1, temp_ptr, 1);
		ptrGPS->satellites[2] = '\0';
	}

	// latitude
	strncpy(ptrGPS->rawLatitude, nWord(ptr, 2), 9);
	ptrGPS->latitude = wordToRad(nWord(ptr, 2));

	// longitude
	strncpy(ptrGPS->rawLongitude, nWord(ptr, 4), 10);
	ptrGPS->longitude  = wordToRad(nWord(ptr, 4));

	if(ptrGPS->speed > topSpeed)
	{
		topSpeed = ptrGPS->speed;
	}
}

char* nWord(char* ptr, uint8_t n)
{
	uint8_t index = 0;
	uint8_t counts = 0;

	for(index = 0; ptr[index] != '\0'; index++)
	{
		if(ptr[index] == ',')
		{
			counts++;

			if(counts == n)
			{
				return(&ptr[index+1]);
			}
		}					
	}

	return(ptr);
}

float wordToFloat(char *ptr)
{
	float value = 0.0;
    int16_t tmp = 0;
	int16_t mag = 0;
        
	while(isdigit(ptr[0]))
	{
    	tmp *= 10;
        tmp += ptr[0] - '0';
        ptr++;
    }

    value = (float)tmp;

    if(ptr[0] == '.')
    {
    	tmp = 0;
        mag = 1;
        
		while(isdigit((++ptr)[0]))
        {
        	tmp *= 10;
            tmp += ptr[0] - '0';
            mag *= 10;
        }
        
		value += (float)tmp / (float)mag;
	}
    
	return(value);
}


float wordToRad(char* ptr)
{
	float angle;
	int16_t tmp;
	int16_t mag;
	
	// wrong address given 
	if(ptr[0] == ',' || ptr[0] == '\0')
	{
		return(0.0);
	}
	
	// decimal point or comma?
	ptr = strchr(ptr, '.');
	if(!ptr)
	{
		ptr = strchr(ptr, ',');
	}

	// conversion 
	tmp = ((ptr[-4] - '0') * 600) +
	      ((ptr[-3] - '0') * 60) +
	      ((ptr[-2] - '0') * 10) +
	      ((ptr[-1] - '0') * 1);

	if(isdigit(ptr[-5]))
	{
		tmp += (ptr[-5] - '0') * 6000;
	}

	angle = (float)tmp;

	if(*ptr == '.')
	{
		tmp = 0;
		mag = 1;

		while(*(++ptr) != ',')
		{
			tmp *= 10;
			tmp += *ptr - '0';
			mag *= 10;
		}

		angle += (float)tmp / (float)mag;
	}

	if(ptr[1] == 'S' || ptr[1] == 'W')
	{
		angle = -angle;
	}

	// convert from minutes to radians 
	angle *= 0.0002908882087;

	return(angle);
}

float wordToTime(char* ptr)
{
	float time;
	int32_t tmp;
	int16_t mag;
	
	// wrong address given 
	if(ptr[0] == ',' || ptr[0] == '\0')
	{
		return(0.0);
	}
	
	// normalize to seconds
	tmp = ((ptr[0] - '0') * 36000) +
	      ((ptr[1] - '0') * 3600) +
	      ((ptr[2] - '0') * 600) +
	      ((ptr[3] - '0') * 60) +
	      ((ptr[4] - '0') * 10) +
	      ((ptr[5] - '0') * 1);

	time = (float)tmp;
	ptr += 6;

	if(*ptr == '.')
	{
		tmp = 0;
		mag = 1;

		while(*(++ptr) != ',')
		{
			tmp *= 10;
			tmp += *ptr - '0';
			mag *= 10;
		}

		time += (float)tmp / (float)mag;
	}

	return(time);
}

/*************************************** USART ***********************************************/
void initUSART(void)
{
	USART_DDR |= (1 << USART_TX_PIN); // tx pin as output 
	USART_DDR &= ~(1 << USART_RX_PIN); // rx pin as input 

	// disable receiver and transmitter
	UCSRB &= ~(1 << RXEN) | ~(1 << TXEN);

	// set baud rate
	UBRRH = (unsigned char)(UBRR_VALUE >> 8);
	UBRRL = (unsigned char)(UBRR_VALUE & 0xFF);

	UCSRA |= (1 << U2X); // double transmitter speed

	// enable receiver and transmitter
	UCSRB |= (1 << RXEN) | (1 << TXEN);

	// asynchron 8N1 
	UCSRC |= (1 << URSEL) | (1 << UCSZ1) | (1 << UCSZ0);
}

/******** sends data over usart ********/
void USART_Send(const char *data) 
{
	while(*data != '\0') // so lange *data != '\0', kein String-Endzeichen enthlt
	{
		while(!(UCSRA & (1 << UDRE))) // Warten bis Senden mglich ist
		{
		}
		
		UDR = *data; // Pointer auf character
		data++; // Nchsten character schnappen
	}
}

/******** read byte from usart ********/
char USART_Read_Byte(void)
{
	while(!(UCSRA & (1 << RXC))); // wait till char. available

    return(UDR); // return char.
}

uint8_t USART_recvACK(char* pattern, uint8_t size)
{
	char data = ' ';
	uint8_t index = 0;
	uint8_t recvMsg = 0;

	// catch all incoming data
	for(index = 0; index < 80; index++)
	{
		data = USART_Read_Byte(); // read and ...

		// message begin
		if(data == '$')
		{
			recvMsg = 1;
		}

		// end of the message reached?
		else if(data == '\r')
		{
			recvMsg = 0;
		}

		if(recvMsg)
		{
			if(data	!= pattern[index])
			{
				return(0);
			}
		}
	}

	return(1);
}

uint8_t USART_isValidGPS(char* pattern, uint8_t size)
{
	char data = ' ';
	uint8_t index = 0;

	// get all messages and check them for an valid flag in the
	// RMC message status 
	data = USART_Read_Byte();

	// message begin
	if(data == '$')
	{
		// catch all incoming data
		for(index = 0; index < 80; index++)
		{
			data = USART_Read_Byte(); // read and ...

			// end of the message reached?
			if(data == '\r')
			{
				// mark the end of the message
				buffer[index] = '\0';

				// is message is a RMC message?
				if(strncmp(&buffer[2], pattern, size) == 0)
				{
					// yep it is, get the char @ status pos.
					data = *nWord(buffer, 2);

					// check the status of the RMC message
					if(data == 'A')
					{
						return(1); // GPS data vaild, leave function
					}
	
					index = 0; // catch a new message
					break;
				}	
			}

			else
			{
				buffer[index] = data; // ... write to buffer
			}
		}
	}

	return(0);
}

/*************************************** Timer1 ***********************************************/
void initTimer1(void)
{
	// timer 1
	TCCR1A = 0; // No Compare Output Mode or Waveform Generation
	TCCR1B = (1 << CS11); // Vorteiler 8 -> 4MHz / 8 = 500kHz Timer Frequenz 
	TCNT1 = -5000; // Timer Register wird mit einem Wert vorbelegt, 2^16 - 60536 = 5000  
}

/*************************************** LCD ***********************************************/
/******** LCD initialisieren ********/
void LCD_init(void)
{
	LCD_DDR = LCD_DDR | 0x0F | (1 << LCD_RS) | (1 << LCD_EN) | (1 << LCD_BACKGRND); // Pins auf Ausgang schalten

	_delay_ms(15);
	LCD_PORT = (LCD_PORT & 0xF0) | 0x0C;
           
	LCD_PORT &= ~(1 << LCD_RS); // Register Select auf "Transferring Instruction Data" setzen
   	LCD_enable();
 
   	_delay_ms(5);
   	LCD_enable();
 
   	_delay_ms(1);
   	LCD_enable();
   	_delay_ms(1);
 
   	// 4 Bit Modus aktivieren 
	LCD_PORT = (LCD_PORT & 0xF0) | 0x04;
	LCD_enable();
   	_delay_ms(1);
 
   	LCD_command(0x28); // 4Bit / 2 Zeilen / 5x7
    LCD_command(0x0C); // Display ein / Cursor aus / kein Blinken
 	LCD_command(0x06); // inkrement / kein Scrollen

   	LCD_clear();
	LCD_home();
}

/******** erzeugt den Enable-Puls ********/
void LCD_enable(void)
{	
	LCD_PORT |= (1 << LCD_EN); // LCD_ENABLED = 1 an LCD Port senden
    _delay_us(2); // kurze Pause zum verarbeiten des Befehls
  	LCD_PORT &= ~(1 << LCD_EN); // LCD_ENABLED = 0 an LCD Port senden
}

/******** Befehl an das LCD senden ********/
void LCD_command(unsigned char temp1)
{
	unsigned char temp2 = temp1;
 
	LCD_PORT &= ~(1 << LCD_RS); // Register Select auf "Transfering Instruction Data" setzen
 
	temp1 = (temp1 >> 4) & 0x0F; // Oberes Nibble holen + Maskieren
	LCD_PORT = (LCD_PORT & 0xF0) | temp1; // Daten Bits setzen
	LCD_enable();
 
	temp2 = temp2 & 0x0F; // unteres Nibble holen und maskieren
	LCD_PORT = (LCD_PORT & 0xF0) | temp2; // Daten Bits setzen
   	LCD_enable();
   
	_delay_us(42);
}

/******** LCD lschen ********/
void LCD_clear(void)
{
	LCD_command(CLEAR_DISPLAY);
   	_delay_ms(5);
}

/******** LCD Cursor Home ********/
void LCD_home(void)
{
	LCD_command(CURSOR_HOME);
   	_delay_ms(5);
}

/******** Datenbyte an das LCD senden ********/
void LCD_data(unsigned char temp1)
{
	unsigned char temp2 = temp1;
 
	LCD_PORT |= (1 << LCD_RS); // Register Select auf "Transferring Display Data" setzen
 
	temp1 = temp1 >> 4;
   	temp1 = temp1 & 0x0F;
	LCD_PORT = (LCD_PORT & 0xF0) | temp1; // Daten Bits setzen
   	
   	LCD_enable();

   	temp2 = temp2 & 0x0F;
	LCD_PORT = (LCD_PORT & 0xF0) | temp2; // Daten Bits setzen
   	
   	LCD_enable();
   
   	_delay_us(42);
}

/******** setzt Cursor in Zeile y (1..4) Spalte x (0..20) ********/
void LCD_set_cursor(uint8_t x, uint8_t y)
{
	uint8_t tmp;
 
  	switch(y) 
	{
    	case 1: // 1. Zeile 
			tmp = 0x80 + 0x00 + x; 
		break;

    	case 2: // 2. Zeile 
			tmp = 0x80 + 0x40 + x; 
		break;

		case 3: // 3. Zeile
            tmp = 0x80 + 0x14 + x;
        break;
 
        case 4: // 4. Zeile
            tmp = 0x80 + 0x54 + x;
        break;

		default: 
			return; // fr den Fall einer falschen Zeile
  	}

  	LCD_command(tmp);
}

/******** display intro ********/
void LCD_intro(void)
{
	uint8_t index = 0;

	LCD_clear();
	LCD_set_cursor(7, 1);

	// "DefCon"
	strcpy_P(cBuffer, INFO_1);
	LCD_string(cBuffer);

	LCD_set_cursor(4, 3);

	// "GPS Laptimer"
	strcpy_P(cBuffer, INFO_2);
	LCD_string(cBuffer);	

	LCD_set_cursor(6, 4);
	LCD_string("ver. ");
	LCD_string(STR(VERSION));

	LCD_set_cursor(0, 2);

	for(index = 0; index < 20; index++)
	{
		LCD_data('.');
		_delay_ms(200);
	}
}

/******** schreibt String auf LCD ********/
void LCD_string(const char *data)
{
	while(*data != '\0') 
	{
    	LCD_data(*data++);
    }
}

void LCD_showTime(struct myTime* time)
{	
	if(time->min < 10) LCD_data('0');

	// print minutes
	LCD_string(utoa(time->min, cBuffer, 10));	
	LCD_data(':');

	if(time->sec < 10) LCD_data('0');

	// print seconds
	LCD_string(utoa(time->sec, cBuffer, 10));
	LCD_data('.');	

	if(time->csec < 10) LCD_data('0');

	// print centiseconds
	LCD_string(utoa(time->csec, cBuffer, 10));
}

void LCD_showLaptime(void)
{
	LCD_home();
	LCD_string("ACT:");
	LCD_set_cursor(12, 1); // Cursor auf Position setzen
  	LCD_showTime(&show);
	
	LCD_set_cursor(0, 2);
	LCD_string("BEST:");
	LCD_set_cursor(12, 2); // Cursor auf Position setzen
  	LCD_showTime(&best);

	LCD_set_cursor(0, 3);
	LCD_string("SPEED:");

	LCD_set_cursor(0, 4);
	LCD_string("DISTANCE:");
}

void LCD_showUTCTime(struct myUTCTime* time)
{
	if(time->hour < 10) LCD_data('0');

	// print hours
	LCD_string(itoa(time->hour, cBuffer, 10));	
	LCD_data(':');

	if(time->min < 10) LCD_data('0');

	// print minutes
	LCD_string(itoa(time->min, cBuffer, 10));	
	LCD_data(':');

	if(time->sec < 10) LCD_data('0');

	// print seconds
	LCD_string(itoa(time->sec, cBuffer, 10));
	
	LCD_data('.');

	if(time->csec < 10) LCD_data('0');
	 	
	// print centiseconds	
	LCD_string(itoa(time->csec, cBuffer, 10));
}

void LCD_showSpeed(struct GPSData *ptrGPS)
{
	if(ptrGPS->speed < 100) LCD_data(' ');
	if(ptrGPS->speed < 10) LCD_data(' ');
	LCD_string(utoa((uint8_t)ptrGPS->speed, cBuffer, 10));	
	LCD_string(" km/h");
}

void LCD_ctrlBacklight(uint8_t on_off)
{
	if(on_off == TRUE)
	{
		LCD_PORT |= (1 << LCD_BACKGRND);
	}

	else
	{
		LCD_PORT &= ~(1 << LCD_BACKGRND);
	}
}

#ifdef DEBUG
void LCD_showDebug(struct GPSData *ptrGPS)
{
	LCD_home();	
	LCD_string("LAT.:");

	if(isdigit(ptrGPS->rawLatitude[0]))
	{
		LCD_set_cursor(11, 1);
		LCD_string(ptrGPS->rawLatitude); // latitude
	}

	else
	{
		LCD_set_cursor(11, 1);
		LCD_string("      N/A"); // not available	
	}

	LCD_set_cursor(0, 2);
	LCD_string("LONG.:");

	if(isdigit(ptrGPS->rawLongitude[0]))
	{
		LCD_set_cursor(10, 2);
		LCD_string(ptrGPS->rawLongitude); // longitude
	}

	else
	{
		LCD_set_cursor(10, 2);
		LCD_string("       N/A"); // not available
	}

	LCD_set_cursor(0, 3);
	LCD_string("SPEED:");
	LCD_set_cursor(12, 3);
	LCD_showSpeed(ptrGPS); // speed
			
	LCD_set_cursor(0, 4);
	LCD_string("HEADING:");
	LCD_set_cursor(9, 4);	
	LCD_string(&headingText[((int)(actGPSPos.heading) * 2 / 45) * 4]);	// heading

	LCD_set_cursor(13, 4);
	LCD_string("SAT:");
	LCD_set_cursor(18, 4);	
	LCD_string(ptrGPS->satellites);	// satellites used
}
#endif


